/*
     tplib.c 

     Copyright (C) 2002 Christian Troger <ct@aim-sw.com>

     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
     as published by the Free Software Foundation; either version 2
     of the License, or (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <sys/time.h>
#include <unistd.h>
#include "tplib.h"

struct message {
	struct timeval timestamp;
	unsigned short length;
	unsigned char data[64];
};

int tp_eib_write_telegram (int device, L_Data_Frame ld)
{
	int i;
	struct message tel;

	tel.data[0] = ld.L_Service | ld.priority | ld.repeated;
	tel.data[1] = ld.source_addr / 256;
	tel.data[2] = ld.source_addr % 256;
	tel.data[3] = ld.dest_addr / 256;
	tel.data[4] = ld.dest_addr % 256;
	memcpy (&tel.data[5], ld.data, ld.datalength);
	tel.length = ld.datalength + 6;

	/* calc checksum */
	tel.data[tel.length - 1] = 0;
        for (i=0;i<(tel.length-1);i++)
		tel.data[tel.length - 1] = 
			((tel.data[i] & 0xff) ^ tel.data[tel.length - 1]);
	tel.data[tel.length - 1] = ~tel.data[tel.length - 1];;

	if (write (device, &tel, sizeof (tel)) <= 0)
		return TPLIB_ERR_DEVICE;
	
	return TPLIB_OK;
}

int tp_eib_extract_L_Data_Frame (unsigned short length, unsigned char data[64], L_Data_Frame *ld)
{
	/* L_Service contains repeated and class-flags! */
	ld->L_Service = data[0];
	if (length > 1) {
  	  ld->priority = (data[0] & L_PRIORITY_MASK);
	  ld->repeated = (data[0] & L_REPEAT_MASK);
	  ld->source_addr = (data[1] * 256 + data[2]);
	  ld->dest_addr = (data[3] * 256 + data[4]);
	  ld->datalength = (length - 6); /* Do not copy checksum */
	  memcpy (ld->data, &data[5], ld->datalength);
	 }

	return TPLIB_OK;
}

