/*
     XTPUART
     Copyright (C) 2002 Christian Troger <ct@aim-sw.com>

     Based on XEIBDump
     Copyright (C) 2001 Bernd Thallner <bernd@kangaroo.at>

     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
     as published by the Free Software Foundation; either version 2
     of the License, or (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <time.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <gtk/gtk.h>
#include <assert.h>

#include "tplib.h"
#include "tpuart_main.h"

#define TPUART_DEVICE "/dev/tpuart0"

#define GUI_FRAME_BORDER 10
#define MAX_TP_MESSAGES 64
#define MAX_DESCRIPTION_LEN 25
#define FILE_MAGIC_CHAR 17

struct message {
  struct timespec timestamp;
  unsigned short length;
  unsigned char data[64];
};

struct tp_message {
  unsigned char description[MAX_DESCRIPTION_LEN];
  int pause;
  L_Data_Frame l_data;
};
  
static char *xtpuart_version="0.01";

GtkWidget *text, *tophbox, *separator, *grledit, *framemsg, *listboxgrl;
GtkWidget *radio1, *radio2, *listboxmsg;
GtkWidget *edent1, *edent2, *edent21, *edent3, *edent4, *edent5;
GtkWidget *priocombo;
int tpuart_fd=0;
gint tag;
int logenabled = 1;
int recv_msg_count=0;
unsigned short tpuart_addr = 0x0001;
char *argv0;
struct tp_message tp_message_list[MAX_TP_MESSAGES];

void dump(char *str, int type) {
  GdkColormap *cmap;
  GdkColor color;
  char time_str[64];
  time_t t;
  struct tm *local_time;

  if (logenabled == 0)
    return;

  time(&t);
  local_time=localtime(&t);
  color.red=0;
  color.green=0;
  color.blue=0;
  cmap=gdk_colormap_get_system();

  switch(type) {
  case 1:
    color.blue=0x9000;
    break;
    
  case 2:
    color.green=0x5000;
    break;

  case 3:
    color.red=0x9000;
    break;
  }

  if(!gdk_color_alloc(cmap, &color)) {
    g_error("couldn't allocate color");
  }
  snprintf(time_str, 64, "%02d:%02d:%02d  ", local_time->tm_hour, local_time->tm_min, local_time->tm_sec);
  gtk_text_insert(GTK_TEXT(text), NULL, &color, NULL, time_str, -1);
  gtk_text_insert(GTK_TEXT(text), NULL, &color, NULL, str, -1);
}

void dialog(char *message) {
  GtkWidget *window, *label, *button;

  window=gtk_dialog_new();
  gtk_window_set_title(GTK_WINDOW (window), "Error");
  gtk_widget_set_usize(window, 400, 100);
  button=gtk_button_new_with_label("  OK  ");
  gtk_signal_connect_object(GTK_OBJECT(button), "clicked", (GtkSignalFunc)gtk_widget_destroy, GTK_OBJECT(window));
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(window)->action_area), button, FALSE, FALSE, 0);
  gtk_widget_show(button);
  label=gtk_label_new(message);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(window)->vbox), label, TRUE, TRUE, 0);
  gtk_widget_show(label);
  gtk_widget_show(window);
}

void _clear_log (void) {
  gtk_text_set_point(GTK_TEXT(text), 0);
  gtk_text_forward_delete(GTK_TEXT(text), gtk_text_get_length(GTK_TEXT(text)));
}

static void clear_window(GtkWidget *w, gpointer data) {
  _clear_log();
}

static void enable_log(gpointer callback_data, guint callback_action, GtkWidget *menu_item) {
  if (GTK_CHECK_MENU_ITEM(menu_item)->active)
    logenabled = 1;
  else {
    logenabled = 0;
    _clear_log();
  }
}

static void reset_device(GtkWidget *w, gpointer data) {
  ioctl (tpuart_fd, TPUART_RESET);
  dump("TPUART Reset OK\n", 2);
}

void _set_group_address (unsigned short ad) {
  char dmp[64];

  ioctl (tpuart_fd, TPUART_SET_GR_ADDR, &ad);
  snprintf(dmp,64,"Group address 0x%04x set\n", ad);
  dump(dmp, 2);
}

void input_callback(gpointer data, gint source, GdkInputCondition condition) {
  int i,len;
  struct message telegram;
  unsigned char msg[5*64];
  L_Data_Frame ldf;

  int __dmpCB (unsigned char v) {
    if ((ldf.L_Service & v) == v)
      return 1;
    else
      return 0;
  }

  if ((len=read(tpuart_fd, &telegram, sizeof (telegram))) <= 0) 
    dump("Cannot read from TPUART\n", 3);
  else
  if (len) {
    tp_eib_extract_L_Data_Frame (telegram.length, telegram.data, &ldf);

    dump ("-------------------------------------------------\n", 0);
    snprintf (msg, 5*64, "Seconds: %lu\n", (unsigned long)telegram.timestamp.tv_sec);
    dump (msg, 0);
    snprintf (msg, 5*64, "Nanoseconds: %lu\n", (unsigned long)telegram.timestamp.tv_nsec);
    dump (msg, 0);
    snprintf (msg, 5*64, "Wall time: %s", ctime (&telegram.timestamp.tv_sec));
    dump (msg, 0);
    snprintf (msg, 5*64, "Data length: %u\n", telegram.length);
    dump (msg, 0);
    snprintf (msg, 5*64, "Count: %u\n", recv_msg_count++);
    dump (msg, 0);
    snprintf (msg, 5*64, "Control byte: 0x%02x ", ldf.L_Service);
    if (__dmpCB (L_DATA_REQ))      snprintf (&msg[19], 5*64, "(L_DATA_REQ)\n"); else
    if (__dmpCB (L_DATA_REQ_LONG)) snprintf (&msg[19], 5*64, "(L_DATA_REQ_LONG)\n"); else
    if (__dmpCB (L_POLL_DATA_REQ)) snprintf (&msg[19], 5*64, "(L_POLL_DATA_REQ)\n"); else
    if (__dmpCB (L_DATA_CONF))     snprintf (&msg[19], 5*64, "(L_DATA_CONF)\n"); else
    if (__dmpCB (L_DATA_NO_CONF))  snprintf (&msg[19], 5*64, "(L_DATA_NO_CONF)\n"); else
    if (__dmpCB (L_ACKFRAME_ACK))  snprintf (&msg[19], 5*64, "(L_ACKFRAME_ACK)\n"); else
    if (__dmpCB (L_ACKFRAME_NAK))  snprintf (&msg[19], 5*64, "(L_ACKFRAME_NAK)\n"); else
    if (__dmpCB (L_ACKFRAME_BUSY)) snprintf (&msg[19], 5*64, "(L_ACKFRAME_BUSY)\n"); else
                                   snprintf (&msg[19], 5*64, "(Unknown)\n");
                                 
    dump (msg, 0);
    if (telegram.length > 1) {
      switch (ldf.priority) {
        case L_PRIORITY_CLASS_SYSTEM: snprintf(msg, 5*64, "Priority: System\n");
  	break;
        case L_PRIORITY_CLASS_ALARM: snprintf(msg, 5*64, "Priority: Alarm\n");
  	break;
        case L_PRIORITY_CLASS_HIGH: snprintf(msg, 5*64, "Priority: High\n");
  	break;
        case L_PRIORITY_CLASS_LOW: snprintf(msg, 5*64, "Priority: Low\n");
  	break;
        default: snprintf(msg, 5*64, "Priority: (Unknown)\n");
  	break;
      }
      dump (msg, 0);
      if (ldf.repeated == L_NREPEAT_DATA)
        snprintf (msg, 5*64, "Repeated: No\n");
      else
        snprintf (msg, 5*64, "Repeated: Yes\n");
      dump (msg, 0);
      snprintf (msg, 5*64, "Source address: 0x%04x\n", ldf.source_addr);
      dump (msg, 0);
      snprintf (msg, 5*64, "Destination address: 0x%04x\n", ldf.dest_addr);
      dump (msg, 0);
  
      if (ldf.datalength > 0)  {
        if ((ldf.data[0] & L_DEST_ADDR_FLAG) == L_DEST_ADDR_FLAG)
          snprintf (msg, 5*64, "Is group address: Yes\n");
        else
          snprintf (msg, 5*64, "Is group address: No\n");
        dump (msg, 0);
      }
      snprintf (msg, 5*64, "Data: ");
      for (i = 0; i < ldf.datalength; i++)
        snprintf (&msg[i*5+6], 6, "0x%02x ", ldf.data[i]);
      msg[ldf.datalength * 5 + 5] = '\n';
      dump (msg, 0);
    } /* if telegram.length > 1 */
  }
}

void quit(GtkWidget *widget, gpointer data) {

  ioctl (tpuart_fd, TPUART_RESET);
  dump("TPUART Reset OK\n", 2);
  gdk_input_remove(tag);
  close(tpuart_fd);
  gtk_main_quit();
}

void save_log_as_ok(GtkWidget *widget, GtkFileSelection *file) {
  char *dump;
  int fd, len;

  dump=gtk_editable_get_chars(GTK_EDITABLE(text), 0, len=gtk_text_get_length(GTK_TEXT(text)));
  if((fd=open(gtk_file_selection_get_filename(GTK_FILE_SELECTION(file)), O_WRONLY | O_CREAT | O_TRUNC , S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH))==-1) {
    dialog(strerror(errno));
    return;
  }

  if(write(fd, dump, len)==-1) {
    dialog(strerror(errno));
    return;
  }
  if(close(fd)==-1) {
    dialog(strerror(errno));
    return;
  }
  free(dump);
  gtk_widget_destroy(GTK_WIDGET(file));
}

static void new_config(GtkWidget *w, gpointer data); //fwd

#define _TMLI tp_message_list[i]

void save_config_ok(GtkWidget *widget, GtkFileSelection *file) {
  FILE *fp;
  int i,m;
  gchar *grl;
  char dmp[64];

  void __putint (int v) {
    snprintf (dmp, 64, "%d\n", v);
    fputs (dmp, fp);
    return;
  }

  if ((fp=fopen(gtk_file_selection_get_filename(GTK_FILE_SELECTION(file)),"w"))==NULL) {
    dialog(strerror(errno));
    return;
  }
  
  fputc (FILE_MAGIC_CHAR, fp);

  /* Group addresses */
  fputc (GTK_CLIST(listboxgrl)->rows, fp);
  for (i=0; i< GTK_CLIST(listboxgrl)->rows; i++) {
    gtk_clist_get_text (GTK_CLIST(listboxgrl), i, 0, &grl);
    fputs (grl, fp);
    fputc ('\n', fp);
  }

  /* Messages */
  fputc (GTK_CLIST(listboxmsg)->rows, fp);
  for (i=0; i< GTK_CLIST(listboxmsg)->rows; i++) {
    fputs (_TMLI.description, fp);
    fputc ('\n', fp);
    __putint (_TMLI.pause);
    fputc (_TMLI.l_data.L_Service,fp);
    fputc (_TMLI.l_data.priority,fp);
    fputc (_TMLI.l_data.repeated,fp);
    __putint (_TMLI.l_data.source_addr);
    __putint (_TMLI.l_data.dest_addr);
    fputc (_TMLI.l_data.datalength,fp);
    for (m=0;m<_TMLI.l_data.datalength;m++)
      fputc (_TMLI.l_data.data[m],fp);
  }

  if(fclose(fp)==EOF) {
    dialog(strerror(errno));
    return;
  }
  gtk_widget_destroy(GTK_WIDGET(file));
}

void load_config_ok(GtkWidget *widget, GtkFileSelection *file) {
  FILE *fp;
  gchar msg[1024];
  gchar *grl;
  int cnt,i,m;
  unsigned short tmp;

  int __getint (void) {
    fgets (msg, 1023, fp);
    msg[strlen(msg)-1]='\0';
    return strtol (msg, NULL, 10);
  }

  if ((fp=fopen(gtk_file_selection_get_filename(GTK_FILE_SELECTION(file)),"r"))==NULL) {
    dialog(strerror(errno));
    return;
  }

  if (fgetc(fp) != FILE_MAGIC_CHAR)
    dialog("Not a valid XTPUART-config file");
  else {
    new_config(NULL,NULL);

    /* Group addresses */
    cnt = fgetc (fp);
    for (i=0; i< cnt; i++) {
      fgets (msg, 1023, fp);
      /* Overwrite the \n we manually added while saving */
      msg [strlen (msg)-1] = '\0';
      grl = msg;
      gtk_clist_append (GTK_CLIST(listboxgrl), &grl);
      tmp = strtol (msg, NULL, 16);
      if (tmp != 0)
        _set_group_address (tmp);
    }

    /* Messages */
    cnt = fgetc (fp); 
    for (i=0; i< cnt; i++) {
      fgets (msg, 1023, fp);
      /* Overwrite the \n we manually added while saving */
      msg [strlen (msg)-1] = '\0';
      strcpy (_TMLI.description, msg);
      grl = msg;
      gtk_clist_append (GTK_CLIST(listboxmsg), &grl);
      _TMLI.pause = __getint ();
      _TMLI.l_data.L_Service = fgetc(fp);
      _TMLI.l_data.priority = fgetc(fp);
      _TMLI.l_data.repeated = fgetc(fp);
      _TMLI.l_data.source_addr = __getint ();
      _TMLI.l_data.dest_addr = __getint ();
      _TMLI.l_data.datalength = fgetc(fp);
      for (m=0;m<_TMLI.l_data.datalength;m++)
        _TMLI.l_data.data[m] = fgetc(fp);
    }
  }

  if(fclose(fp)==EOF) {
    dialog(strerror(errno));
    return;
  }
  gtk_widget_destroy(GTK_WIDGET(file));
}

static void save_log_as(GtkWidget *w, gpointer data) {
  GtkWidget *file;

  file=gtk_file_selection_new ("Save Log As");
  gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION (file)->ok_button), "clicked", (GtkSignalFunc)save_log_as_ok, file);
  gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(file)), "destroy", (GtkSignalFunc)gtk_widget_destroy, GTK_OBJECT(file));
  gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(file)->cancel_button), "clicked", (GtkSignalFunc)gtk_widget_destroy, GTK_OBJECT(file));
  gtk_file_selection_set_filename(GTK_FILE_SELECTION(file), "tpuart.log");
  gtk_widget_show(file);
}

static void save_config_as(GtkWidget *w, gpointer data) {
  GtkWidget *file;

  file=gtk_file_selection_new ("Save Configuration As");
  gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION (file)->ok_button), "clicked", (GtkSignalFunc)save_config_ok, file);
  gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(file)), "destroy", (GtkSignalFunc)gtk_widget_destroy, GTK_OBJECT(file));
  gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(file)->cancel_button), "clicked", (GtkSignalFunc)gtk_widget_destroy, GTK_OBJECT(file));
  gtk_file_selection_set_filename(GTK_FILE_SELECTION(file), "config.xtp");
  gtk_widget_show(file);
}

static void load_config_as(GtkWidget *w, gpointer data) {
  GtkWidget *file;

  file=gtk_file_selection_new ("Load Configuration");
  gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION (file)->ok_button), "clicked", (GtkSignalFunc)load_config_ok, file);
  gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(file)), "destroy", (GtkSignalFunc)gtk_widget_destroy, GTK_OBJECT(file));
  gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(file)->cancel_button), "clicked", (GtkSignalFunc)gtk_widget_destroy, GTK_OBJECT(file));
  gtk_file_selection_set_filename(GTK_FILE_SELECTION(file), "config.xtp");
  gtk_widget_show(file);
}

static void grl_removeall_click(GtkWidget *w, gpointer data); //fwd
static void msg_removeall_click(GtkWidget *w, gpointer data); //fwd
static void msg_clear_click(GtkWidget *w, gpointer data); //fwd

static void new_config(GtkWidget *w, gpointer data) {
  grl_removeall_click(w, data);
  msg_removeall_click(w, data);
  msg_clear_click(w, data);
}

void about(GtkWidget *widget, gpointer data) {
  GtkWidget *window, *text, *button;
  GdkFont *font;
  GdkColormap *cmap;
  GdkColor color;
  char str[512];

  window=gtk_dialog_new();
  gtk_window_set_title(GTK_WINDOW(window), "About");
  gtk_widget_set_usize(window, 400, 240);
  button=gtk_button_new_with_label("  OK  ");
  gtk_signal_connect_object(GTK_OBJECT(button), "clicked", (GtkSignalFunc)gtk_widget_destroy, GTK_OBJECT(window));
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(window)->action_area), button, FALSE, FALSE, 0);
  gtk_widget_show(button);

  text=gtk_text_new(NULL, NULL);
  gtk_text_set_editable(GTK_TEXT(text), FALSE);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(window)->vbox), text, TRUE, TRUE, 0);
  gtk_widget_show(text);

  cmap=gdk_colormap_get_system();
  color.red=0;
  color.green=0;
  color.blue=0xffff;
  if(!gdk_color_alloc(cmap, &color)) {
    g_error("couldn't allocate color");
  }

  font=gdk_font_load("-*-times-medium-i-normal-*-*-300-*-*-*-*-*-*");
  snprintf(str, 64, " XTPUART %s\n", xtpuart_version);
  gtk_text_insert(GTK_TEXT(text), font, &color, NULL, str, -1);
  snprintf(str, 512, "\n XTPUART version %s\n Copyright (C) 2002 Christian Troger <ct@aim-sw.com>\n\n Based on XEIBDump version 0.2\n Copyright (C) 2001 Bernd Thallner <bernd@kangaroo.at>\n\n This software comes with ABSOLUTELY NO WARRANTY;\n for details see COPYING file.", xtpuart_version);
  gtk_text_insert(GTK_TEXT(text), NULL, NULL, NULL, str, -1);

  gtk_widget_show(window);
}

static GtkItemFactoryEntry menu_items[]={
  { "/_File",                        NULL,         NULL,                   0, "<Branch>" },
  { "/File/_New Configuration",      NULL,	   new_config,             0, NULL },
  { "/File/_Load Configuration",     NULL,         load_config_as,         0, NULL },
  { "/File/_Save Configuration",     NULL,         save_config_as,         0, NULL },
  { "/File/separator",               NULL,         NULL,                   0, "<Separator>" },
  { "/File/Quit",                    "<control>Q", quit,                   0, NULL },
  { "/_Device",                      NULL,         NULL,                   0, "<Branch>"},
  { "/Device/_Reset TPUART",   	     "<control>R", reset_device,           0, NULL },
  { "/_Log",                         NULL,         NULL,                   0, "<Branch>"},
  { "/Log/_Clear",       	     "<control>C", clear_window,           0, NULL },
  { "/Log/_Save As...",              "<control>S", save_log_as,		   0, NULL },
  { "/Log/separator",                NULL,         NULL,                   0, "<Separator>" },
  { "/Log/_Enabled",       	     "<Control>E", enable_log,             1, "<CheckItem>"},
  { "/_Help",                        NULL,         NULL,                   0, "<LastBranch>" },
  { "/Help/_About...",               NULL,         about,                  0, NULL } };


/*************************************************************/
/* Normal/Busmonitor mode                                    */

static void normal_mode_click(GtkWidget *w, gpointer data) {

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(radio1))) {
    gtk_widget_show (framemsg);
    gtk_widget_show (separator);

    /* TPUART_BUSMON_OFF is exactly the same as TPUART_RESET */
    ioctl (tpuart_fd, TPUART_BUSMON_OFF);
    dump("Busmonitor-Mode deactivated\n",2);
  }
}

static void busmon_mode_click(GtkWidget *w, gpointer data) {

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(radio2))) {
    gtk_widget_hide (framemsg);
    gtk_widget_hide (separator);
    ioctl (tpuart_fd, TPUART_RESET);
    dump("TPUART Reset OK\n",2);
    ioctl (tpuart_fd, TPUART_BUSMON_ON);
    dump("Busmonitor-Mode activated\n",2);
  }
}

/*************************************************************/
/* Button Signals GRL                                        */

static void grl_add_click(GtkWidget *w, gpointer data)  {
  gchar *grl;
  gchar msg[7];
  unsigned short tmp;

  tmp = strtol (gtk_entry_get_text (GTK_ENTRY (grledit)), NULL, 16);

  if (tmp != 0) {
    _set_group_address(tmp);
    snprintf (msg, 7, "0x%04x", tmp);
    grl = msg;
    gtk_clist_append (GTK_CLIST(listboxgrl), &grl);
    gtk_entry_set_text (GTK_ENTRY (grledit), "");
  }
  else
  dialog ("Group address 0x0000 not allowed (=Broadcast)");
}

void _remove_tpuart_gr_addr (gint row) {
  unsigned short tmp;
  gchar *grl;
  char dmp[64];

  gtk_clist_get_text (GTK_CLIST(listboxgrl), row, 0, &grl);
  tmp = strtol (grl, NULL, 16);
  ioctl (tpuart_fd, TPUART_UNSET_GR_ADDR, &tmp);
  snprintf(dmp,64,"Group address 0x%04x cleared\n", tmp);
  dump(dmp, 2);
}

static void grl_remove_click(GtkWidget *w, gpointer data)  {
  gint selrow;

  if (GTK_CLIST(listboxgrl)->selection == NULL)
    selrow = -1;
  else
    selrow = (gint)(GTK_CLIST(listboxgrl)->selection->data);

  if (selrow < 0)
    dialog("No group address selected");
  else {
    _remove_tpuart_gr_addr (selrow);
    gtk_clist_remove(GTK_CLIST(listboxgrl), selrow);
  }
}

static void grl_removeall_click(GtkWidget *w, gpointer data)  {
  int i;

  for (i=0;i<(GTK_CLIST(listboxgrl)->rows);i++)
    _remove_tpuart_gr_addr (i);

  gtk_clist_clear (GTK_CLIST(listboxgrl));
}

/*************************************************************/
/* Button Signals MSG                                        */

#define _LBM GTK_CLIST(listboxmsg)

int _find_descr (gchar *des) {
  int i;
  gchar *tmp;

  for (i=0;i<(_LBM->rows);i++) {
    gtk_clist_get_text (_LBM, i, 0, &tmp);
    if (strcmp (tmp, des) == 0)
      return i;
  }
  return -1;
}

void _set_priority_data (unsigned char prio) {

  switch (prio) {
    case L_PRIORITY_CLASS_SYSTEM:
      gtk_entry_set_text (GTK_ENTRY(GTK_COMBO(priocombo)->entry), "System");
      break;
    case L_PRIORITY_CLASS_ALARM:
      gtk_entry_set_text (GTK_ENTRY(GTK_COMBO(priocombo)->entry), "Alarm");
      break;
    case L_PRIORITY_CLASS_HIGH:
      gtk_entry_set_text (GTK_ENTRY(GTK_COMBO(priocombo)->entry), "High");
      break;
    case L_PRIORITY_CLASS_LOW:
    default:
      gtk_entry_set_text (GTK_ENTRY(GTK_COMBO(priocombo)->entry), "Low");
      break;
  }

}

unsigned char _get_priority_data(void) {
  gchar *pr;

  pr = gtk_entry_get_text (GTK_ENTRY(GTK_COMBO(priocombo)->entry));
  if (strcmp (pr, "System") == 0)
    return L_PRIORITY_CLASS_SYSTEM;
  else
    if (strcmp (pr, "Alarm") == 0)
      return L_PRIORITY_CLASS_ALARM;
    else
      if (strcmp (pr, "High") == 0)
        return L_PRIORITY_CLASS_HIGH;
      else
        return L_PRIORITY_CLASS_LOW;
}

unsigned char _fill_APP_data (int idx) {
  gchar *dt;
  unsigned int x[15];
  int cnt,i;

  dt = gtk_entry_get_text (GTK_ENTRY(edent3));
  cnt = sscanf (dt,"%02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x",
                   &x[0],&x[1],&x[2],&x[3],&x[4],&x[5],&x[6],&x[7],&x[8],&x[9],
                   &x[10],&x[11],&x[12],&x[13],&x[14]);
  if (cnt < 0)
    cnt = 0;

  for (i=0;i<cnt;i++)
    tp_message_list[idx].l_data.data[2+i] = x[i];
  return cnt;
}

void _rewrite_APP_data (int idx) {
  int cnt,i;
  gchar dt[15*3+1];
 
  cnt = tp_message_list[idx].l_data.datalength - 2;

  for (i=0;i<cnt;i++)
    snprintf (&dt[i*3], 4, "%02x ", tp_message_list[idx].l_data.data[2+i]);
  if (cnt != 0)
    dt[cnt*3-1] = '\0';
  else
    strcpy (dt,"");
  gtk_entry_set_text (GTK_ENTRY(edent3), dt);
}

int _getrow_check (void) {
  gint selrow;

  if (_LBM->selection == NULL)
    selrow = -1;
  else
    selrow = (gint)(_LBM->selection->data);

  if (selrow < 0)
    dialog("No message selected");
  return selrow;
}

void _do_edit (int addnew) {
  gchar *des;
  int i;
  int fd;
  unsigned char lng = 0;

  if (((_LBM->rows) >= MAX_TP_MESSAGES) && (addnew == 1))
    dialog("Maximum number of messages reached");
  else {
    des = gtk_entry_get_text(GTK_ENTRY(edent1));
    fd = _find_descr (des);
    if (strcmp (des,"") == 0)
      dialog("Message with empty description not allowed");
    else 
      if ((fd != -1) && (addnew == 1))
        dialog("A message with this description already exists");
      else {
        if (addnew == 1) { 
          gtk_clist_append (_LBM,&des);
          i = (GTK_CLIST(listboxmsg)->rows) - 1;
          assert (i>=0);
        }
        else {
          if ((i = _getrow_check ()) == -1)
            return;
          if ((fd != i) && (fd != -1)) {
            dialog("A message with this description already exists");
            return;
          }
          gtk_clist_set_text (GTK_CLIST(listboxmsg), i, 0, des);
        }

        /* Description */
        strcpy (tp_message_list [i].description, des);

        /* Pause */
        tp_message_list[i].pause = 
          strtol (gtk_entry_get_text (GTK_ENTRY (edent4)), NULL, 10);

        /* Control Field + Source + Dest */
        tp_message_list[i].l_data.L_Service = L_DATA_REQ;
        tp_message_list[i].l_data.priority = _get_priority_data();
        tp_message_list[i].l_data.repeated = L_NREPEAT_DATA;
        tp_message_list[i].l_data.source_addr = tpuart_addr;
        tp_message_list[i].l_data.dest_addr =
          strtol (gtk_entry_get_text (GTK_ENTRY (edent2)), NULL, 16);

        /* Octet 6 */
        tp_message_list[i].l_data.data[1] = 
          strtol (gtk_entry_get_text (GTK_ENTRY (edent5)), NULL, 16);

        /* APP Data */
        lng = _fill_APP_data (i);
        tp_message_list[i].l_data.datalength = lng + 2; 

        /* Octet 5 */
        if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (edent21)))
          tp_message_list[i].l_data.data[0] = L_DEST_ADDR_FLAG + lng;
        else
          tp_message_list[i].l_data.data[0] = lng;
      }
  }
}

void _clear_data (void) {
  /* Description */
  gtk_entry_set_text(GTK_ENTRY(edent1), "");

  /* Pause */
  gtk_entry_set_text (GTK_ENTRY (edent4), "0");

  /* Dest */
  gtk_entry_set_text (GTK_ENTRY (edent2), "0000");

  /* Priority */
  _set_priority_data(L_PRIORITY_CLASS_LOW);

  /* Octet 6 */
  gtk_entry_set_text (GTK_ENTRY (edent5), "00");

  /* APP Data */
  gtk_entry_set_text (GTK_ENTRY (edent3), "");

  /* Octet 5 */
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (edent21), FALSE);
}

static void msg_add_click(GtkWidget *w, gpointer data)  {
  _do_edit (1);
}

static void msg_edit_click(GtkWidget *w, gpointer data)  {
  _do_edit (0);
}

static void msg_remove_click(GtkWidget *w, gpointer data)  {
  gint selrow = _getrow_check();
  int i;

  if (selrow != -1) {
    for (i = selrow; i < (_LBM->rows-1); i++)
      tp_message_list[i] = tp_message_list[i+1];

    gtk_clist_remove(_LBM, selrow);
  }
}

static void msg_removeall_click(GtkWidget *w, gpointer data)  {
  gtk_clist_clear (_LBM);
}


static void msg_clear_click(GtkWidget *w, gpointer data)  {
  _clear_data();
}

static void msg_movedown_click(GtkWidget *w, gpointer data)  {
  int selrow = _getrow_check ();
  struct tp_message msg;
 
  if ((selrow != -1) && (selrow < (_LBM->rows-1))) {
    gtk_clist_swap_rows (_LBM, selrow, selrow+1);
    msg = tp_message_list[selrow];
    tp_message_list[selrow] = tp_message_list[selrow+1];
    tp_message_list[selrow+1] = msg;
  }
}

static void msg_moveup_click(GtkWidget *w, gpointer data)  {
  int selrow = _getrow_check ();
  struct tp_message msg;
 
  if (selrow > 0) {
    gtk_clist_swap_rows (_LBM, selrow, selrow-1);
    msg = tp_message_list[selrow];
    tp_message_list[selrow] = tp_message_list[selrow-1];
    tp_message_list[selrow-1] = msg;
  }
}

void msg_select_row_callback (GtkWidget *widget, gint row, gint column, GdkEventButton *event, gpointer data) {
  gchar msg[7];

  /* Description */
  gtk_entry_set_text(GTK_ENTRY(edent1), tp_message_list[row].description);

  /* Pause */
  snprintf (msg, 7, "%d", tp_message_list[row].pause);
  gtk_entry_set_text (GTK_ENTRY (edent4), msg);

  /* Dest */
  snprintf (msg, 7, "%04x", tp_message_list[row].l_data.dest_addr);
  gtk_entry_set_text (GTK_ENTRY (edent2), msg);
  _set_priority_data(tp_message_list[row].l_data.priority);

  /* Octet 6 */
  snprintf (msg, 7, "%02x", tp_message_list[row].l_data.data[1]);
  gtk_entry_set_text (GTK_ENTRY (edent5), msg);

  /* APP Data */
  _rewrite_APP_data (row);

  /* Octet 5 */
  if ((tp_message_list[row].l_data.data[0] & L_DEST_ADDR_FLAG) == L_DEST_ADDR_FLAG)
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (edent21), TRUE);
  else
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (edent21), FALSE);
}

void _do_send (int idx) {
  char tmp[256];

  snprintf (tmp, 256, "Sending message '%s'\n", tp_message_list[idx].description);
  dump (tmp,0);

  if (tp_eib_write_telegram (tpuart_fd, tp_message_list[idx].l_data) != TPLIB_OK)
    dump ("Cannot write message to TPUART\n", 3);
}

static void msg_send_click(GtkWidget *w, gpointer data)  {
  int selrow = _getrow_check ();
  
  if (selrow != -1) 
    _do_send (selrow);
}

static void msg_sendall_click(GtkWidget *w, gpointer data)  {
  int i;
  char tmp[256];

  for (i=0;i<_LBM->rows;i++) {
    _do_send (i);
    snprintf (tmp, 256, "Pausing %d ms\n", tp_message_list[i].pause);
    dump (tmp,0);
    usleep (tp_message_list[i].pause * 1000);
  }
  dump ("All messages sent\n",2);
}

/*************************************************************/
/* MAIN                                                      */

int main(int argc, char *argv[]) {
  GtkWidget *window, *scrollbar, *hbox, *vbox, *menubar;
  GtkWidget *framelog, *topswbox, *separator2;
  GtkWidget *sboxgrl, *framegrl, *edboxmsg, *btboxmsg, *btboxgrl, *btgrl1,*btgrl2, *btgrl3, *addgroup;
  GtkWidget *sboxmsg, *btmsg1,*btmsg2, *btmsg3, *btmsg4, *btmsg5, *btmsg6, *btmsg7, *btmsg8, *btmsg9, *btmsg10;
 GtkWidget *scrboxgrl, *scrboxmsg;
 GtkWidget *bttablemsg,*edtablemsg,*bttablegrl;
 GtkWidget *edlab1, *edlab2, *edlab3, *edlab4, *edlab5, *edlabP;
 GList *combolist = NULL;

  GtkItemFactory *item_factory;
  GtkAccelGroup *accel_group;
  gint nmenu_items=sizeof(menu_items)/sizeof(menu_items[0]);
  char *tpuart_device=TPUART_DEVICE, c, str[64];

  gchar *grlheader = "Group addr.";
  gchar *msgheader = "Messages                         ";

  gtk_init(&argc, &argv);
  argv0=argv[0];
  while((c=getopt(argc, argv, "dp:vh"))!=EOF) {
    switch(c) {

    case 'd':
      tpuart_device=optarg;
      break;

    case 'p':
      tpuart_addr=strtol(optarg, NULL, 16); 
      break;

    case 'v':
      fprintf(stdout, "XTPUART %s\n", xtpuart_version);
      return 0;

    default:
    case '?':
    case 'h':
      fprintf(stdout, "XTPUART version %s\n   Copyright (C) 2002 Christian Troger <ct@aim-sw.com>\n", xtpuart_version);
      fprintf(stdout, "Based on XEIBDump version 0.2\n   Copyright (C) 2001 Bernd Thallner <bernd@kangaroo.at>\n");
      fprintf(stdout, "This software comes with ABSOLUTELY NO WARRANTY; for details see COPYING file.\n\n");
      fprintf(stdout, "Usage: %s [options]\nOptions:\n", argv0);
      fprintf(stdout, "  -d <device>       Set TPUART device name\n");
      fprintf(stdout, "  -p <address>      Set TPUART physical address (Base 16)\n");
      fprintf(stdout, "  -v                Print program version number\n");
      fprintf(stdout, "  -h                Print this help\n\n");
      return 0;
    }
  }

  if((tpuart_fd=open(tpuart_device, O_RDWR))==-1) {
    snprintf(str, 64, "%s: %s", argv0, tpuart_device);
    perror(str);
    return 1;
  }

  ioctl (tpuart_fd, TPUART_SET_PH_ADDR, &tpuart_addr);

  window=gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_widget_set_usize(window, 790, 385);
  gtk_window_set_title(GTK_WINDOW (window), "XTPUART");
  gtk_window_set_policy(GTK_WINDOW(window), TRUE, TRUE, FALSE);
  gtk_signal_connect(GTK_OBJECT(window), "destroy", GTK_SIGNAL_FUNC(quit), NULL);

  gtk_container_set_border_width(GTK_CONTAINER(window), 1);

  vbox=gtk_vbox_new(FALSE, 1);
  gtk_container_border_width(GTK_CONTAINER(vbox), 0);
  gtk_container_add(GTK_CONTAINER(window), vbox);
  gtk_widget_show(vbox);

  accel_group=gtk_accel_group_new();
  item_factory = gtk_item_factory_new(GTK_TYPE_MENU_BAR, "<main>", accel_group);
  gtk_item_factory_create_items(item_factory, nmenu_items, menu_items, NULL);
  gtk_window_add_accel_group(GTK_WINDOW(window), accel_group);
  menubar=gtk_item_factory_get_widget(item_factory, "<main>");
  gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(gtk_item_factory_get_item(item_factory, "/Log/Enabled")), TRUE);

  gtk_box_pack_start(GTK_BOX(vbox), menubar, FALSE, TRUE, 0);
  gtk_widget_show(menubar);

  /*************************************************************/
  /* Top Switch Box for Normal/Busmon Mode & Radio Buttons     */
  topswbox=gtk_hbox_new(FALSE,0);
  gtk_box_pack_start(GTK_BOX(vbox), topswbox, FALSE, FALSE, 0);
  gtk_widget_show(topswbox);

  radio1 = gtk_radio_button_new_with_label (NULL, "Normal mode");
  gtk_signal_connect_object(GTK_OBJECT(radio1), "clicked", (GtkSignalFunc)normal_mode_click, GTK_OBJECT(window));
  gtk_box_pack_start (GTK_BOX(topswbox), radio1, FALSE, TRUE, 0);
  gtk_widget_show (radio1);

  radio2 = gtk_radio_button_new_with_label_from_widget(
    GTK_RADIO_BUTTON (radio1), "Busmonitor mode");
  gtk_signal_connect_object(GTK_OBJECT(radio2), "clicked", (GtkSignalFunc)busmon_mode_click, GTK_OBJECT(window));
  gtk_box_pack_start (GTK_BOX(topswbox), radio2, FALSE, TRUE, 0);
  gtk_widget_show (radio2);

  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (radio1), TRUE);
  /* Radio Buttons END */


  /*************************************************************/
  /* Separator                                                 */
  separator = gtk_hseparator_new();
  gtk_box_pack_start (GTK_BOX (vbox), separator, FALSE, TRUE, 0);
  gtk_widget_show (separator);

  /*************************************************************/
  /* Message List                                              */
  framemsg=gtk_frame_new(NULL);
  gtk_frame_set_shadow_type (GTK_FRAME (framemsg), GTK_SHADOW_OUT);
  gtk_container_border_width(GTK_CONTAINER(framemsg), GUI_FRAME_BORDER);

  /* Set the following to FALSE, TRUE for fixed top msg-frame */
  gtk_box_pack_start(GTK_BOX(vbox), framemsg, TRUE, TRUE,0);
  gtk_widget_show(framemsg);

  sboxmsg=gtk_hbox_new(FALSE, 0);
  gtk_container_add(GTK_CONTAINER(framemsg),sboxmsg);
  gtk_widget_show(sboxmsg);

  /* Scrollbox for listbox */
  scrboxmsg=gtk_scrolled_window_new(NULL,NULL);
  gtk_container_add(GTK_CONTAINER(sboxmsg), scrboxmsg);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW(scrboxmsg), GTK_POLICY_NEVER, GTK_POLICY_ALWAYS);
  gtk_widget_show(scrboxmsg);

  /* Listbox for messages */
  listboxmsg = gtk_clist_new_with_titles(1,&msgheader);
  gtk_signal_connect(GTK_OBJECT(listboxmsg), "select_row", GTK_SIGNAL_FUNC(msg_select_row_callback), NULL);
  gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(scrboxmsg), listboxmsg);
  gtk_clist_set_shadow_type(GTK_CLIST(listboxmsg), GTK_SHADOW_NONE);
  gtk_widget_show(listboxmsg);
  
  /* Box for button table */
  btboxmsg = gtk_vbox_new (FALSE, 0);
  gtk_box_pack_start (GTK_BOX(sboxmsg),btboxmsg,FALSE,TRUE,0);
  gtk_widget_show (btboxmsg);
  
  /* Button table */
  bttablemsg = gtk_table_new (6,2,FALSE);
  gtk_box_pack_start (GTK_BOX(btboxmsg), bttablemsg,FALSE,FALSE,0);
  gtk_container_set_border_width(GTK_CONTAINER(bttablemsg), 10);
  gtk_widget_show(bttablemsg);

  /* Buttons */
  btmsg1 = gtk_button_new_with_label(" Add ");
  gtk_signal_connect_object(GTK_OBJECT(btmsg1), "clicked", 
    (GtkSignalFunc)msg_add_click, GTK_OBJECT(window));
  gtk_widget_show(btmsg1);
  btmsg2 = gtk_button_new_with_label(" Remove ");
  gtk_signal_connect_object(GTK_OBJECT(btmsg2), "clicked", 
    (GtkSignalFunc)msg_remove_click, GTK_OBJECT(window));
  gtk_widget_show(btmsg2);
  btmsg3 = gtk_button_new_with_label(" Move down ");
  gtk_signal_connect_object(GTK_OBJECT(btmsg3), "clicked", 
    (GtkSignalFunc)msg_movedown_click, GTK_OBJECT(window));
  gtk_widget_show(btmsg3);
  btmsg4 = gtk_button_new_with_label(" Send ");
  gtk_signal_connect_object(GTK_OBJECT(btmsg4), "clicked", 
    (GtkSignalFunc)msg_send_click, GTK_OBJECT(window));
  gtk_widget_show(btmsg4);
  btmsg5 = gtk_button_new_with_label(" Replace ");
  gtk_signal_connect_object(GTK_OBJECT(btmsg5), "clicked", 
    (GtkSignalFunc)msg_edit_click, GTK_OBJECT(window));
  gtk_widget_show(btmsg5);
  btmsg6 = gtk_button_new_with_label(" Remove all ");
  gtk_signal_connect_object(GTK_OBJECT(btmsg6), "clicked", 
    (GtkSignalFunc)msg_removeall_click, GTK_OBJECT(window));
  gtk_widget_show(btmsg6);
  btmsg7 = gtk_button_new_with_label(" Move up ");
  gtk_signal_connect_object(GTK_OBJECT(btmsg7), "clicked", 
    (GtkSignalFunc)msg_moveup_click, GTK_OBJECT(window));
  gtk_widget_show(btmsg7);
  btmsg8 = gtk_button_new_with_label(" Send all ");
  gtk_signal_connect_object(GTK_OBJECT(btmsg8), "clicked", 
    (GtkSignalFunc)msg_sendall_click, GTK_OBJECT(window));
  gtk_widget_show(btmsg8);
  btmsg9 = gtk_button_new_with_label("  ");
//  gtk_signal_connect_object(GTK_OBJECT(btmsg9), "clicked", 
//    (GtkSignalFunc)msg_duplicate_click, GTK_OBJECT(window));
  gtk_widget_show(btmsg9);
  btmsg10 = gtk_button_new_with_label(" Clear ");
  gtk_signal_connect_object(GTK_OBJECT(btmsg10), "clicked", 
    (GtkSignalFunc)msg_clear_click, GTK_OBJECT(window));
  gtk_widget_show(btmsg10);

  gtk_table_attach_defaults (GTK_TABLE(bttablemsg), btmsg1, 0,1,0,1);
  gtk_table_attach_defaults (GTK_TABLE(bttablemsg), btmsg2, 0,1,1,2);
  gtk_table_attach_defaults (GTK_TABLE(bttablemsg), btmsg3, 0,1,2,3);
  gtk_table_attach_defaults (GTK_TABLE(bttablemsg), btmsg9, 0,1,3,4);
  gtk_table_attach_defaults (GTK_TABLE(bttablemsg), btmsg5, 1,2,0,1);
  gtk_table_attach_defaults (GTK_TABLE(bttablemsg), btmsg6, 1,2,1,2);
  gtk_table_attach_defaults (GTK_TABLE(bttablemsg), btmsg7, 1,2,2,3);
  gtk_table_attach_defaults (GTK_TABLE(bttablemsg), btmsg10,1,2,3,4);
  gtk_table_attach_defaults (GTK_TABLE(bttablemsg), btmsg4, 0,1,4,5);
  gtk_table_attach_defaults (GTK_TABLE(bttablemsg), btmsg8, 1,2,4,5);
  /* Message List END */

  /*************************************************************/
  /* EditTable of MessageList                                  */

  /* Box for table */
  edboxmsg = gtk_vbox_new (FALSE, 0);
  gtk_box_pack_start (GTK_BOX(sboxmsg),edboxmsg,FALSE,TRUE,0);
  gtk_widget_show (edboxmsg);
  
  /* Button table */
  edtablemsg = gtk_table_new (4,5,FALSE);
  gtk_box_pack_start (GTK_BOX(edboxmsg), edtablemsg,FALSE,FALSE,0);
  gtk_container_set_border_width(GTK_CONTAINER(edtablemsg), 10);
  gtk_widget_show(edtablemsg);

  /* Labels */
  edlab1 = gtk_label_new("Description ");
  gtk_label_set_justify(GTK_LABEL(edlab1), GTK_JUSTIFY_LEFT);
  gtk_widget_show (edlab1);
  edlab2 = gtk_label_new("Destination ");
  gtk_label_set_justify(GTK_LABEL(edlab2), GTK_JUSTIFY_LEFT);
  gtk_widget_show (edlab2);
  edlab3 = gtk_label_new("Data (O7-) ");
  gtk_label_set_justify(GTK_LABEL(edlab3), GTK_JUSTIFY_LEFT);
  gtk_widget_show (edlab3);
  edlab4 = gtk_label_new(" Pause (ms) ");
  gtk_label_set_justify(GTK_LABEL(edlab4), GTK_JUSTIFY_LEFT);
  gtk_widget_show (edlab4);
  edlab5 = gtk_label_new(" Data (O6)");
  gtk_label_set_justify(GTK_LABEL(edlab5), GTK_JUSTIFY_LEFT);
  gtk_widget_show (edlab5);
  edlabP = gtk_label_new("            Priority");
  gtk_label_set_justify(GTK_LABEL(edlabP), GTK_JUSTIFY_LEFT);
  gtk_widget_show (edlabP);

  /* Entries */
  edent1 = gtk_entry_new();
  gtk_entry_set_max_length (GTK_ENTRY (edent1), MAX_DESCRIPTION_LEN);
  gtk_widget_show (edent1);

  /* Dest */
  edent2 = gtk_entry_new();
  gtk_entry_set_max_length (GTK_ENTRY (edent2), 4);
  gtk_widget_set_usize(edent2, 50,23);
  gtk_widget_show (edent2);

  edent21 = gtk_check_button_new_with_label("Group addr.");
  gtk_widget_show (edent21);

  /* Data */
  edent3 = gtk_entry_new();
  gtk_entry_set_max_length (GTK_ENTRY (edent3), 15 * 3);
  gtk_widget_show (edent3);

  /* Pause */
  edent4 = gtk_entry_new();
  gtk_entry_set_max_length (GTK_ENTRY (edent4), 6);
  gtk_widget_set_usize(edent4, 50,23);
  gtk_widget_show (edent4);
  
  /* Octet 6 */
  edent5 = gtk_entry_new();
  gtk_entry_set_max_length (GTK_ENTRY (edent5), 2);
  gtk_widget_set_usize(edent5, 50,23);
  gtk_widget_show (edent5);

  /* Priority */
  combolist = g_list_append (combolist, "Low");
  combolist = g_list_append (combolist, "High");
  combolist = g_list_append (combolist, "Alarm");
  combolist = g_list_append (combolist, "System");

  priocombo = gtk_combo_new();
  gtk_combo_set_popdown_strings (GTK_COMBO(priocombo), combolist);
  gtk_widget_set_usize(priocombo, 50,23);
  gtk_widget_show (priocombo); 

/*
  gtk_table_attach_defaults (GTK_TABLE(edtablemsg), edlab1, 0,1,0,1); 
  gtk_table_attach_defaults (GTK_TABLE(edtablemsg), edlab2, 0,1,1,2);
  gtk_table_attach_defaults (GTK_TABLE(edtablemsg), edlab3, 2,3,2,3);
  gtk_table_attach_defaults (GTK_TABLE(edtablemsg), edlab4, 3,4,3,4);
  gtk_table_attach_defaults (GTK_TABLE(edtablemsg), edlab5, 0,1,2,3);
  gtk_table_attach_defaults (GTK_TABLE(edtablemsg), edlabP, 0,1,3,4);
  gtk_table_attach_defaults (GTK_TABLE(edtablemsg), edent1, 1,5,0,1);
  gtk_table_attach_defaults (GTK_TABLE(edtablemsg), edent2, 1,2,1,2);
  gtk_table_attach_defaults (GTK_TABLE(edtablemsg), edent21,2,3,1,2);
  gtk_table_attach_defaults (GTK_TABLE(edtablemsg), edent3, 3,5,2,3);
  gtk_table_attach_defaults (GTK_TABLE(edtablemsg), edent4, 4,5,3,4);
  gtk_table_attach_defaults (GTK_TABLE(edtablemsg), edent5, 1,2,2,3);
  gtk_table_attach_defaults (GTK_TABLE(edtablemsg), priocombo, 1,3,3,4);*/

  gtk_table_attach_defaults (GTK_TABLE(edtablemsg), edlab1, 0,1,0,1); 
  gtk_table_attach_defaults (GTK_TABLE(edtablemsg), edlab2, 0,1,1,2);
  gtk_table_attach_defaults (GTK_TABLE(edtablemsg), edlabP, 2,3,2,3);
  gtk_table_attach_defaults (GTK_TABLE(edtablemsg), edlab4, 3,4,0,1);
  gtk_table_attach_defaults (GTK_TABLE(edtablemsg), edlab5, 0,1,2,3);
  gtk_table_attach_defaults (GTK_TABLE(edtablemsg), edlab3, 0,1,3,4);
  gtk_table_attach_defaults (GTK_TABLE(edtablemsg), edent1, 1,3,0,1);
  gtk_table_attach_defaults (GTK_TABLE(edtablemsg), edent2, 1,2,1,2);
  gtk_table_attach_defaults (GTK_TABLE(edtablemsg), edent21,2,3,1,2);
  gtk_table_attach_defaults (GTK_TABLE(edtablemsg), edent3, 1,5,3,4);
  gtk_table_attach_defaults (GTK_TABLE(edtablemsg), edent4, 4,5,0,1);
  gtk_table_attach_defaults (GTK_TABLE(edtablemsg), edent5, 1,2,2,3);
  gtk_table_attach_defaults (GTK_TABLE(edtablemsg), priocombo, 3,5,2,3);

  /*************************************************************/
  /* Separator                                                 */
  separator2 = gtk_hseparator_new();
  gtk_box_pack_start (GTK_BOX (vbox), separator2, FALSE, TRUE, 0);
  gtk_widget_show (separator2);

  /*************************************************************/
  /*  H Box for groupaddresses, log,                           */
  tophbox=gtk_hbox_new(FALSE,0);
  gtk_container_add(GTK_CONTAINER(vbox), tophbox);
  gtk_widget_show(tophbox);
  /* TOP H END */

  /*************************************************************/
  /* Group List                                                */
  framegrl=gtk_frame_new(NULL);
  gtk_frame_set_shadow_type (GTK_FRAME (framegrl), GTK_SHADOW_OUT);
  gtk_container_border_width(GTK_CONTAINER(framegrl), GUI_FRAME_BORDER);
  gtk_box_pack_start(GTK_BOX(tophbox), framegrl, FALSE, TRUE, 0);
  gtk_widget_show(framegrl);

  sboxgrl=gtk_hbox_new(FALSE, 0);
  gtk_container_add(GTK_CONTAINER(framegrl),sboxgrl);
  gtk_widget_show(sboxgrl);

  scrboxgrl=gtk_scrolled_window_new(NULL,NULL);
  gtk_container_add(GTK_CONTAINER(sboxgrl), scrboxgrl);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW(scrboxgrl), GTK_POLICY_NEVER, GTK_POLICY_ALWAYS);
  gtk_widget_show(scrboxgrl);

  listboxgrl = gtk_clist_new_with_titles(1,&grlheader);
  gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(scrboxgrl), listboxgrl);
  gtk_clist_set_shadow_type(GTK_CLIST(listboxgrl), GTK_SHADOW_NONE);
  gtk_widget_show(listboxgrl);
  
  /* Box for button table */
  btboxgrl = gtk_vbox_new(FALSE, 0);
  gtk_box_pack_start (GTK_BOX(sboxgrl),btboxgrl,FALSE,TRUE,0);
  gtk_widget_show(btboxgrl);
  
  /* Button & Edit table */
  bttablegrl = gtk_table_new (3,1,FALSE);
  gtk_box_pack_start (GTK_BOX(btboxgrl), bttablegrl,FALSE,FALSE,0);
  gtk_container_set_border_width(GTK_CONTAINER(bttablegrl), 10);
  gtk_widget_show(bttablegrl);

  addgroup = gtk_hbox_new (FALSE, 0);
  gtk_widget_show(addgroup);

  /* Add/Edit Box */
   grledit = gtk_entry_new();
   gtk_entry_set_max_length (GTK_ENTRY (grledit), 4);
   gtk_widget_set_usize(grledit, 50,23);
   gtk_container_add(GTK_CONTAINER(addgroup),grledit);
   gtk_widget_show (grledit);

   btgrl1 = gtk_button_new_with_label(" Add ");
   gtk_signal_connect_object(GTK_OBJECT(btgrl1), "clicked", 
     (GtkSignalFunc)grl_add_click, GTK_OBJECT(window));
   gtk_container_add(GTK_CONTAINER(addgroup),btgrl1);
   gtk_widget_show(btgrl1);
  /* Add/Edit Box End */

  btgrl2 = gtk_button_new_with_label("Remove");
  gtk_signal_connect_object(GTK_OBJECT(btgrl2), "clicked", 
    (GtkSignalFunc)grl_remove_click, GTK_OBJECT(window));
  gtk_widget_show(btgrl2);
  btgrl3 = gtk_button_new_with_label("Remove all");
  gtk_signal_connect_object(GTK_OBJECT(btgrl3), "clicked", 
    (GtkSignalFunc)grl_removeall_click, GTK_OBJECT(window));
  gtk_widget_show(btgrl3);

  gtk_table_attach_defaults (GTK_TABLE(bttablegrl), addgroup,0,1,0,1);
  gtk_table_attach_defaults (GTK_TABLE(bttablegrl), btgrl2,0,1,1,2);
  gtk_table_attach_defaults (GTK_TABLE(bttablegrl), btgrl3,0,1,2,3);
  /* Group List END */

  /*************************************************************/
  /* LOG                                                       */
  framelog=gtk_frame_new(NULL);
  gtk_frame_set_shadow_type (GTK_FRAME (framelog), GTK_SHADOW_OUT);
  gtk_container_border_width(GTK_CONTAINER(framelog), GUI_FRAME_BORDER);
  gtk_container_add(GTK_CONTAINER(tophbox), framelog);
  gtk_widget_show(framelog);

  hbox=gtk_hbox_new(FALSE, 0);
  gtk_container_add(GTK_CONTAINER(framelog),hbox);
  gtk_widget_show(hbox);

  text=gtk_text_new(NULL, NULL);
  gtk_text_set_editable(GTK_TEXT(text), FALSE);
  gtk_box_pack_start(GTK_BOX(hbox), text, TRUE, TRUE, 0);
  gtk_widget_show(text);

  scrollbar=gtk_vscrollbar_new(GTK_TEXT(text)->vadj);
  gtk_box_pack_start(GTK_BOX(hbox), scrollbar, FALSE, FALSE, 0);
  gtk_widget_show(scrollbar);
  /* LOG END */

  gtk_widget_show(window);

  _clear_data();

  tag=gdk_input_add(tpuart_fd, GDK_INPUT_READ, input_callback, NULL);
  dump("TPUART device opened\n", 1);
  snprintf (str, 64, "TPUART physical address set to 0x%04x\n", tpuart_addr);
  dump(str, 1);

  gtk_main ();
  return 0;
}
